/* classes: src_files */

/*	Copyright (C) 1994 Free Software Foundation, Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this software; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */



#define TERPP(OBJ) (CAR (OBJ) == scm_tc16_tcl_interp)
#define TERP(OBJ) ((Tcl_Interp *)CDR (OBJ))
#define SCMDATA(OBJ) ((SCM)(TERP(OBJ)->client_data))
#define PROPS(OBJ) (CDR (SCMDATA (OBJ)))
#define SELF(OBJ) (CAR (SCMDATA (OBJ)))
#define SELF_interp(INTERP) (CAR ((SCM)((INTERP)->client_data)))

extern int scm_tc16_tcl_interp;
