# missed symbols from the first version of this script:
s/\bbig2dbl\b/scm_big2dbl/g
s/\bdbl2big\b/scm_dbl2big/g
s/\baddbig\b/scm_addbig/g
s/\badjbig\b/scm_adjbig/g
s/\bbig2inum\b/scm_big2inum/g
s/\bbigcomp\b/scm_bigcomp/g
s/\bcopybig\b/scm_copybig/g
s/\bdivbigbig\b/scm_divbigbig/g
s/\bdivbigdig\b/scm_divbigdig/g
s/\bdivbigint\b/scm_divbigint/g
s/\blong2big\b/scm_long2big/g
s/\bmkbig\b/scm_mkbig/g
s/\bmulbig\b/scm_mulbig/g
s/\bnormbig\b/scm_normbig/g
s/\bpseudolong\b/scm_pseudolong/g
s/\bs_adjbig\b/scm_s_adjbig/g
s/\bs_bignum\b/scm_s_bignum/g
s/\bulong2big\b/scm_ulong2big/g
s/\baind\b/scm_aind/g
s/\baref\b/scm_aref/g
s/\barray2list\b/scm_array2list/g
s/\barray_contents\b/scm_array_contents/g
s/\barray_copy\b/scm_array_copy/g
s/\barray_dims\b/scm_array_dims/g
s/\barray_fill\b/scm_array_fill/g
s/\barray_for_each\b/scm_array_for_each/g
s/\barray_imap\b/scm_array_imap/g
s/\barray_inbp\b/scm_array_inbp/g
s/\barray_map\b/scm_array_map/g
s/\barray_prot\b/scm_array_prot/g
s/\barray_rank\b/scm_array_rank/g
s/\barray_ref\b/scm_array_ref/g
s/\barrayp\b/scm_arrayp/g
s/\baset\b/scm_aset/g
s/\bbit_count\b/scm_bit_count/g
s/\bbit_inv\b/scm_bit_inv/g
s/\bbit_set\b/scm_bit_set/g
s/\bcvref\b/scm_cvref/g
s/\bdims2ura\b/scm_dims2ura/g
s/\bencl_array\b/scm_encl_array/g
s/\blcount\b/scm_lcount/g
s/\blist2ura\b/scm_list2ura/g
s/\bmake_ra\b/scm_make_ra/g
s/\bmake_sh_array\b/scm_make_sh_array/g
s/\bmake_uve\b/scm_make_uve/g
s/\bposition\b/scm_position/g
s/\bra2contig\b/scm_ra2contig/g
s/\bra_difference\b/scm_ra_difference/g
s/\bra_divide\b/scm_ra_divide/g
s/\bra_eqp\b/scm_ra_eqp/g
s/\bra_greqp\b/scm_ra_greqp/g
s/\bra_grp\b/scm_ra_grp/g
s/\bra_leqp\b/scm_ra_leqp/g
s/\bra_lessp\b/scm_ra_lessp/g
s/\bra_matchp\b/scm_ra_matchp/g
s/\bra_product\b/scm_ra_product/g
s/\bra_set_contp\b/scm_ra_set_contp/g
s/\bra_sum\b/scm_ra_sum/g
s/\braequal\b/scm_raequal/g
s/\brafill\b/scm_rafill/g
s/\bramapc\b/scm_ramapc/g
s/\bs_make_sh_array\b/scm_s_make_sh_array/g
s/\bshap2ra\b/scm_shap2ra/g
s/\bstrdown\b/scm_strdown/g
s/\bstrup\b/scm_strup/g
s/\btc16_array\b/scm_tc16_array/g
s/\btrans_array\b/scm_trans_array/g
s/\bura_read\b/scm_ura_read/g
s/\bura_write\b/scm_ura_write/g
s/\buve_len\b/scm_uve_len/g
s/\btc16_variable\b/scm_tc16_variable/g
s/\bapply\b/scm_apply/g
s/\bbadargsp\b/scm_badargsp/g
s/\bceval\b/scm_ceval/g
s/\bclosure\b/scm_closure/g
s/\bcopytree\b/scm_copytree/g
s/\bdefinedp\b/scm_definedp/g
s/\beval\b/scm_eval/g
s/\beval_args\b/scm_eval_args/g
s/\bfor_each\b/scm_for_each/g
s/\bforce\b/scm_force/g
s/\bi_apply\b/scm_i_apply/g
s/\bi_arrow\b/scm_i_arrow/g
s/\bi_dot\b/scm_i_dot/g
s/\bi_else\b/scm_i_else/g
s/\bi_lambda\b/scm_i_lambda/g
s/\bi_let\b/scm_i_let/g
s/\bi_quasiquote\b/scm_i_quasiquote/g
s/\bi_quote\b/scm_i_quote/g
s/\bi_unquote\b/scm_i_unquote/g
s/\bi_uq_splicing\b/scm_i_uq_splicing/g
s/\bilookup\b/scm_ilookup/g
s/\binit_eval\b/scm_init_eval/g
s/\blookupcar\b/scm_lookupcar/g
s/\bm_and\b/scm_m_and/g
s/\bm_apply\b/scm_m_apply/g
s/\bm_begin\b/scm_m_begin/g
s/\bm_case\b/scm_m_case/g
s/\bm_cond\b/scm_m_cond/g
s/\bm_cont\b/scm_m_cont/g
s/\bm_define\b/scm_m_define/g
s/\bm_delay\b/scm_m_delay/g
s/\bm_do\b/scm_m_do/g
s/\bm_if\b/scm_m_if/g
s/\bm_lambda\b/scm_m_lambda/g
s/\bm_let\b/scm_m_let/g
s/\bm_letrec\b/scm_m_letrec/g
s/\bm_letstar\b/scm_m_letstar/g
s/\bm_or\b/scm_m_or/g
s/\bm_quasiquote\b/scm_m_quasiquote/g
s/\bm_quote\b/scm_m_quote/g
s/\bm_set\b/scm_m_set/g
s/\bmakacro\b/scm_makacro/g
s/\bmake_synt\b/scm_make_synt/g
s/\bmakmacro\b/scm_makmacro/g
s/\bmakmmacro\b/scm_makmmacro/g
s/\bmakprom\b/scm_makprom/g
s/\bmap\b/scm_map/g
s/\bnconc2last\b/scm_nconc2last/g
s/\bprocedurep\b/scm_procedurep/g
s/\bs_for_each\b/scm_s_for_each/g
s/\bs_force\b/scm_s_force/g
s/\bs_map\b/scm_s_map/g
s/\btc16_macro\b/scm_tc16_macro/g
s/\btc16_promise\b/scm_tc16_promise/g
s/\bdir_free\b/scm_dir_free/g
s/\bdir_print\b/scm_dir_print/g
s/\bfile_position\b/scm_file_position/g
s/\bfile_set_position\b/scm_file_set_position/g
s/\binit_ioext\b/scm_init_ioext/g
s/\bl_access\b/scm_l_access/g
s/\bl_chmod\b/scm_l_chmod/g
s/\bl_closedir\b/scm_l_closedir/g
s/\bl_dup\b/scm_l_dup/g
s/\bl_dup2\b/scm_l_dup2/g
s/\bl_fileno\b/scm_l_fileno/g
s/\bl_getcwd\b/scm_l_getcwd/g
s/\bl_getpid\b/scm_l_getpid/g
s/\bl_isatty\b/scm_l_isatty/g
s/\bl_mkdir\b/scm_l_mkdir/g
s/\bl_opendir\b/scm_l_opendir/g
s/\bl_readdir\b/scm_l_readdir/g
s/\bl_rewinddir\b/scm_l_rewinddir/g
s/\bl_rmdir\b/scm_l_rmdir/g
s/\bl_stat\b/scm_l_stat/g
s/\bl_umask\b/scm_l_umask/g
s/\bl_utime\b/scm_l_utime/g
s/\blchdir\b/scm_lchdir/g
s/\bopen_pipe\b/scm_open_pipe/g
s/\bread_line\b/scm_read_line/g
s/\bread_line1\b/scm_read_line1/g
s/\bren_fil\b/scm_ren_fil/g
s/\breopen_file\b/scm_reopen_file/g
s/\bs_op_pipe\b/scm_s_op_pipe/g
s/\bs_stat\b/scm_s_stat/g
s/\bstat2scm\b/scm_stat2scm/g
s/\btc16_dir\b/scm_tc16_dir/g
s/\bgrinfo\b/scm_grinfo/g
s/\binit_posix\b/scm_init_posix/g
s/\bl_chown\b/scm_l_chown/g
s/\bl_getegid\b/scm_l_getegid/g
s/\bl_geteuid\b/scm_l_geteuid/g
s/\bl_getgid\b/scm_l_getgid/g
s/\bl_getppid\b/scm_l_getppid/g
s/\bl_getuid\b/scm_l_getuid/g
s/\bl_kill\b/scm_l_kill/g
s/\bl_link\b/scm_l_link/g
s/\bl_pipe\b/scm_l_pipe/g
s/\bl_setegid\b/scm_l_setegid/g
s/\bl_seteuid\b/scm_l_seteuid/g
s/\bl_setgid\b/scm_l_setgid/g
s/\bl_setuid\b/scm_l_setuid/g
s/\bl_ttyname\b/scm_l_ttyname/g
s/\bl_waitpid\b/scm_l_waitpid/g
s/\bpwinfo\b/scm_pwinfo/g
s/\binit_record\b/scm_init_record/g
s/\bloc_makrtd\b/scm_loc_makrtd/g
s/\bmakrectyp\b/scm_makrectyp/g
s/\brec_accessor\b/scm_rec_accessor/g
s/\brec_accessor1\b/scm_rec_accessor1/g
s/\brec_constr\b/scm_rec_constr/g
s/\brec_constr1\b/scm_rec_constr1/g
s/\brec_modifier\b/scm_rec_modifier/g
s/\brec_modifier1\b/scm_rec_modifier1/g
s/\brec_pred\b/scm_rec_pred/g
s/\brec_pred1\b/scm_rec_pred1/g
s/\brec_rtd\b/scm_rec_rtd/g
s/\brecequal\b/scm_recequal/g
s/\brecordp\b/scm_recordp/g
s/\btc16_record\b/scm_tc16_record/g
s/\babrt\b/scm_abrt/g
s/\badd_feature\b/scm_add_feature/g
s/\balrm_deferred\b/scm_alrm_deferred/g
s/\bcells_allocated\b/scm_cells_allocated/g
s/\bchar_readyp\b/scm_char_readyp/g
s/\bcharnames\b/scm_charnames/g
s/\bcharnums\b/scm_charnums/g
s/\bcur_error_port\b/scm_cur_error_port/g
s/\bcur_input_port\b/scm_cur_input_port/g
s/\bcur_output_port\b/scm_cur_output_port/g
s/\bdisplay\b/scm_display/g
s/\bdowncase\b/scm_downcase/g
s/\beof_objectp\b/scm_eof_objectp/g
s/\berr_env\b/scm_err_env/g
s/\berr_exp\b/scm_err_exp/g
s/\berr_pos\b/scm_err_pos/g
s/\berr_s_subr\b/scm_err_s_subr/g
s/\berrjmp_bad\b/scm_errjmp_bad/g
s/\berrmsgs\b/scm_errmsgs/g
s/\beverr\b/scm_everr/g
s/\bexecargv\b/scm_execargv/g
s/\bexit_report\b/scm_exit_report/g
s/\bexitval\b/scm_exitval/g
s/\bfeatures\b/scm_features/g
s/\bfinal_repl\b/scm_final_repl/g
s/\bgc_cells_collected\b/scm_gc_cells_collected/g
s/\bgc_end\b/scm_gc_end/g
s/\bgc_malloc_collected\b/scm_gc_malloc_collected/g
s/\bgc_ports_collected\b/scm_gc_ports_collected/g
s/\bgc_rt\b/scm_gc_rt/g
s/\bgc_start\b/scm_gc_start/g
s/\bgc_time_taken\b/scm_gc_time_taken/g
s/\bgrow_tok_buf\b/scm_grow_tok_buf/g
s/\bgrowth_mon\b/scm_growth_mon/g
s/\bhan_alrm\b/scm_han_alrm/g
s/\bhan_sig\b/scm_han_sig/g
s/\bhandle_it\b/scm_handle_it/g
s/\bheap_report\b/scm_heap_report/g
s/\binit_features\b/scm_init_features/g
s/\binit_init\b/scm_init_init/g
s/\binit_repl\b/scm_init_repl/g
s/\binit_tables\b/scm_init_tables/g
s/\bintprint\b/scm_intprint/g
s/\bints_disabled\b/scm_ints_disabled/g
s/\biprin1\b/scm_iprin1/g
s/\biprlist\b/scm_iprlist/g
s/\bipruk\b/scm_ipruk/g
s/\bisymnames\b/scm_isymnames/g
s/\blcells_allocated\b/scm_lcells_allocated/g
s/\blerrno\b/scm_lerrno/g
s/\blexec\b/scm_lexec/g
s/\blfflush\b/scm_lfflush/g
s/\blflush\b/scm_lflush/g
s/\blfwrite\b/scm_lfwrite/g
s/\blgetc\b/scm_lgetc/g
s/\bline_num\b/scm_line_num/g
s/\blinum\b/scm_linum/g
s/\blmallocated\b/scm_lmallocated/g
s/\bloc_errobj\b/scm_loc_errobj/g
s/\bloc_features\b/scm_loc_features/g
s/\bloc_loadpath\b/scm_loc_loadpath/g
s/\blowers\b/scm_lowers/g
s/\blperror\b/scm_lperror/g
s/\blputc\b/scm_lputc/g
s/\blputs\b/scm_lputs/g
s/\blread\b/scm_lread/g
s/\blroom\b/scm_lroom/g
s/\blungetc\b/scm_lungetc/g
s/\blwrite\b/scm_lwrite/g
s/\bmakarb\b/scm_makarb/g
s/\bmallocated\b/scm_mallocated/g
s/\bnewline\b/scm_newline/g
s/\bpeek_char\b/scm_peek_char/g
s/\bprog_args\b/scm_prog_args/g
s/\bprolixity\b/scm_prolixity/g
s/\bquit\b/scm_quit/g
s/\bread_char\b/scm_read_char/g
s/\brelarb\b/scm_relarb/g
s/\brepl\b/scm_repl/g
s/\brepl_driver\b/scm_repl_driver/g
s/\brepl_report\b/scm_repl_report/g
s/\brestart\b/scm_restart/g
s/\brt\b/scm_rt/g
s/\bs_ccl\b/scm_s_ccl/g
s/\bs_cur_errp\b/scm_s_cur_errp/g
s/\bs_cur_inp\b/scm_s_cur_inp/g
s/\bs_cur_outp\b/scm_s_cur_outp/g
s/\bs_exec\b/scm_s_exec/g
s/\bs_newline\b/scm_s_newline/g
s/\bs_read\b/scm_s_read/g
s/\bs_relarb\b/scm_s_relarb/g
s/\bs_tryarb\b/scm_s_tryarb/g
s/\bs_write\b/scm_s_write/g
s/\bset_errp\b/scm_set_errp/g
s/\bset_inp\b/scm_set_inp/g
s/\bset_outp\b/scm_set_outp/g
s/\bsig_deferred\b/scm_sig_deferred/g
s/\btc16_arbiter\b/scm_tc16_arbiter/g
s/\btmp_errobj\b/scm_tmp_errobj/g
s/\btmp_loadpath\b/scm_tmp_loadpath/g
s/\btrans\b/scm_trans/g
s/\btrans_off\b/scm_trans_off/g
s/\btrans_on\b/scm_trans_on/g
s/\btryarb\b/scm_tryarb/g
s/\btryload\b/scm_tryload/g
s/\bupcase\b/scm_upcase/g
s/\buppers\b/scm_uppers/g
s/\bverbose\b/scm_verbose/g
s/\bwarn\b/scm_warn/g
s/\bwrite_char\b/scm_write_char/g
s/\bwta\b/scm_wta/g
s/\bappendb\b/scm_appendb/g
s/\binit_sc2\b/scm_init_sc2/g
s/\blast_pair\b/scm_last_pair/g
s/\bstrnullp\b/scm_strnullp/g
s/\bsubfl\b/scm_subfl/g
s/\bsubml\b/scm_subml/g
s/\bsubmr\b/scm_submr/g
s/\bangle\b/scm_angle/g
s/\bassv\b/scm_assv/g
s/\bbigequal\b/scm_bigequal/g
s/\bbigprint\b/scm_bigprint/g
s/\bdblprec\b/scm_dblprec/g
s/\bdifference\b/scm_difference/g
s/\bdivide\b/scm_divide/g
s/\beqp\b/scm_eqp/g
s/\bequal\b/scm_equal/g
s/\beqv\b/scm_eqv/g
s/\bexpt\b/scm_expt/g
s/\bfloequal\b/scm_floequal/g
s/\bfloident\b/scm_floident/g
s/\bfloprint\b/scm_floprint/g
s/\bgreaterp\b/scm_greaterp/g
s/\bgreqp\b/scm_greqp/g
s/\bhash\b/scm_hash/g
s/\bhasher\b/scm_hasher/g
s/\bhashq\b/scm_hashq/g
s/\bhashv\b/scm_hashv/g
s/\biint2str\b/scm_iint2str/g
s/\bimag_part\b/scm_imag_part/g
s/\bin2ex\b/scm_in2ex/g
s/\binexactp\b/scm_inexactp/g
s/\binit_scl\b/scm_init_scl/g
s/\bintp\b/scm_intp/g
s/\bistr2flo\b/scm_istr2flo/g
s/\bistr2int\b/scm_istr2int/g
s/\bistring2number\b/scm_istring2number/g
s/\blacosh\b/scm_lacosh/g
s/\blasinh\b/scm_lasinh/g
s/\blatan2\b/scm_latan2/g
s/\blatanh\b/scm_latanh/g
s/\bleqp\b/scm_leqp/g
s/\blessp\b/scm_lessp/g
s/\blgetenv\b/scm_lgetenv/g
s/\blist_tail\b/scm_list_tail/g
s/\blmax\b/scm_lmax/g
s/\blmin\b/scm_lmin/g
s/\blsystem\b/scm_lsystem/g
s/\bltrunc\b/scm_ltrunc/g
s/\bmagnitude\b/scm_magnitude/g
s/\bmakdbl\b/scm_makdbl/g
s/\bmakpolar\b/scm_makpolar/g
s/\bmakrect\b/scm_makrect/g
s/\bmemv\b/scm_memv/g
s/\bnegativep\b/scm_negativep/g
s/\bnum2ulong\b/scm_num2ulong/g
s/\bnum_protects\b/scm_num_protects/g
s/\bnumber2string\b/scm_number2string/g
s/\bnumberp\b/scm_numberp/g
s/\bpositivep\b/scm_positivep/g
s/\bproduct\b/scm_product/g
s/\breal_part\b/scm_real_part/g
s/\brealp\b/scm_realp/g
s/\bround\b/scm_round/g
s/\bs_difference\b/scm_s_difference/g
s/\bs_divide\b/scm_s_divide/g
s/\bs_getenv\b/scm_s_getenv/g
s/\bs_inexactp\b/scm_s_inexactp/g
s/\bs_product\b/scm_s_product/g
s/\bs_sum\b/scm_s_sum/g
s/\bs_system\b/scm_s_system/g
s/\bsofttype\b/scm_softtype/g
s/\bstring2list\b/scm_string2list/g
s/\bstring2number\b/scm_string2number/g
s/\bstring_copy\b/scm_string_copy/g
s/\bstring_fill\b/scm_string_fill/g
s/\bsum\b/scm_sum/g
s/\bsys_protects\b/scm_sys_protects/g
s/\btwo_doubles\b/scm_two_doubles/g
s/\bulong2num\b/scm_ulong2num/g
s/\bvector2list\b/scm_vector2list/g
s/\bvector_fill\b/scm_vector_fill/g
s/\bzerop\b/scm_zerop/g
s/\bignore_signals\b/scm_ignore_signals/g
s/\binit_banner\b/scm_init_banner/g
s/\binit_scm\b/scm_init_scm/g
s/\binit_signals\b/scm_init_signals/g
s/\bl_pause\b/scm_l_pause/g
s/\bl_raise\b/scm_l_raise/g
s/\bl_sleep\b/scm_l_sleep/g
s/\blalarm\b/scm_lalarm/g
s/\bloc_tick_signal\b/scm_loc_tick_signal/g
s/\blticks\b/scm_lticks/g
s/\bmain\b/scm_main/g
s/\brestore_signals\b/scm_restore_signals/g
s/\brun_scm\b/scm_run_scm/g
s/\btick_count\b/scm_tick_count/g
s/\btick_signal\b/scm_tick_signal/g
s/\bticken\b/scm_ticken/g
s/\bunignore_signals\b/scm_unignore_signals/g
s/\binit_socket\b/scm_init_socket/g
s/\bl_accept\b/scm_l_accept/g
s/\bl_bind\b/scm_l_bind/g
s/\bl_connect\b/scm_l_connect/g
s/\bl_gethost\b/scm_l_gethost/g
s/\bl_getpeername\b/scm_l_getpeername/g
s/\bl_getsockname\b/scm_l_getsockname/g
s/\bl_listen\b/scm_l_listen/g
s/\bl_socket\b/scm_l_socket/g
s/\babsval\b/scm_absval/g
s/\bappend\b/scm_append/g
s/\bash\b/scm_ash/g
s/\bassoc\b/scm_assoc/g
s/\bassq\b/scm_assq/g
s/\bbitextract\b/scm_bitextract/g
s/\bbooleanp\b/scm_booleanp/g
s/\bchar2int\b/scm_char2int/g
s/\bchar_alphap\b/scm_char_alphap/g
s/\bchar_downcase\b/scm_char_downcase/g
s/\bchar_geqp\b/scm_char_geqp/g
s/\bchar_grp\b/scm_char_grp/g
s/\bchar_leqp\b/scm_char_leqp/g
s/\bchar_lessp\b/scm_char_lessp/g
s/\bchar_lowerp\b/scm_char_lowerp/g
s/\bchar_nump\b/scm_char_nump/g
s/\bchar_upcase\b/scm_char_upcase/g
s/\bchar_upperp\b/scm_char_upperp/g
s/\bchar_whitep\b/scm_char_whitep/g
s/\bcharp\b/scm_charp/g
s/\bchci_eq\b/scm_chci_eq/g
s/\bchci_geqp\b/scm_chci_geqp/g
s/\bchci_grp\b/scm_chci_grp/g
s/\bchci_leqp\b/scm_chci_leqp/g
s/\bchci_lessp\b/scm_chci_lessp/g
s/\bconsp\b/scm_consp/g
s/\beq\b/scm_eq/g
s/\bevenp\b/scm_evenp/g
s/\bexactp\b/scm_exactp/g
s/\bilength\b/scm_ilength/g
s/\bilentab\b/scm_ilentab/g
s/\binit_iprocs\b/scm_init_iprocs/g
s/\binit_subrs\b/scm_init_subrs/g
s/\bint2char\b/scm_int2char/g
s/\bintexpt\b/scm_intexpt/g
s/\bintlength\b/scm_intlength/g
s/\blength\b/scm_length/g
s/\blgcd\b/scm_lgcd/g
s/\blist\b/scm_list/g
s/\blist_ref\b/scm_list_ref/g
s/\blistp\b/scm_listp/g
s/\bllcm\b/scm_llcm/g
s/\blnot\b/scm_lnot/g
s/\blogand\b/scm_logand/g
s/\blogcount\b/scm_logcount/g
s/\blogior\b/scm_logior/g
s/\blognot\b/scm_lognot/g
s/\blogtab\b/scm_logtab/g
s/\blogxor\b/scm_logxor/g
s/\blquotient\b/scm_lquotient/g
s/\blremainder\b/scm_lremainder/g
s/\bmake_string\b/scm_make_string/g
s/\bmake_vector\b/scm_make_vector/g
s/\bmember\b/scm_member/g
s/\bmemq\b/scm_memq/g
s/\bmodulo\b/scm_modulo/g
s/\bnullp\b/scm_nullp/g
s/\boddp\b/scm_oddp/g
s/\breverse\b/scm_reverse/g
s/\bs_list\b/scm_s_list/g
s/\bs_make_string\b/scm_s_make_string/g
s/\bs_make_vector\b/scm_s_make_vector/g
s/\bsetcar\b/scm_setcar/g
s/\bsetcdr\b/scm_setcdr/g
s/\bst_append\b/scm_st_append/g
s/\bst_equal\b/scm_st_equal/g
s/\bst_geqp\b/scm_st_geqp/g
s/\bst_grp\b/scm_st_grp/g
s/\bst_length\b/scm_st_length/g
s/\bst_leqp\b/scm_st_leqp/g
s/\bst_lessp\b/scm_st_lessp/g
s/\bst_ref\b/scm_st_ref/g
s/\bst_set\b/scm_st_set/g
s/\bstci_equal\b/scm_stci_equal/g
s/\bstci_geqp\b/scm_stci_geqp/g
s/\bstci_grp\b/scm_stci_grp/g
s/\bstci_leqp\b/scm_stci_leqp/g
s/\bstci_lessp\b/scm_stci_lessp/g
s/\bstring\b/scm_string/g
s/\bstring2symbol\b/scm_string2symbol/g
s/\bstringp\b/scm_stringp/g
s/\bsubstring\b/scm_substring/g
s/\bsymbol2string\b/scm_symbol2string/g
s/\bsymbolp\b/scm_symbolp/g
s/\bvector\b/scm_vector/g
s/\bvector_length\b/scm_vector_length/g
s/\bvector_ref\b/scm_vector_ref/g
s/\bvector_set\b/scm_vector_set/g
s/\bvectorp\b/scm_vectorp/g
s/\bacons\b/scm_acons/g
s/\badd_final\b/scm_add_final/g
s/\bclose_port\b/scm_close_port/g
s/\bcons\b/scm_cons/g
s/\bcons2\b/scm_cons2/g
s/\bcwis\b/scm_cwis/g
s/\bcwos\b/scm_cwos/g
s/\bdel_fil\b/scm_del_fil/g
s/\bdowinds\b/scm_dowinds/g
s/\bdynthrow\b/scm_dynthrow/g
s/\bdynwind\b/scm_dynwind/g
s/\bequal0\b/scm_equal0/g
s/\bexpmem\b/scm_expmem/g
s/\bfinals\b/scm_finals/g
s/\bfree0\b/scm_free0/g
s/\bfree_storage\b/scm_free_storage/g
s/\bfreelist\b/scm_freelist/g
s/\bgc\b/scm_gc/g
s/\bgc_for_newcell\b/scm_gc_for_newcell/g
s/\bgc_mark\b/scm_gc_mark/g
s/\bgrew_lim\b/scm_grew_lim/g
s/\bheap_org\b/scm_heap_org/g
s/\bheap_size\b/scm_heap_size/g
s/\bhplim_ind\b/scm_hplim_ind/g
s/\bhplims\b/scm_hplims/g
s/\bigc\b/scm_igc/g
s/\binit_heap_seg\b/scm_init_heap_seg/g
s/\binit_io\b/scm_init_io/g
s/\binit_storage\b/scm_init_storage/g
s/\binit_types\b/scm_init_types/g
s/\binput_portp\b/scm_input_portp/g
s/\bintern\b/scm_intern/g
s/\blthrow\b/scm_lthrow/g
s/\bltmpnam\b/scm_ltmpnam/g
s/\bmakcclo\b/scm_makcclo/g
s/\bmake_cont\b/scm_make_cont/g
s/\bmake_subr\b/scm_make_subr/g
s/\bmakfromstr\b/scm_makfromstr/g
s/\bmakstr\b/scm_makstr/g
s/\bmark0\b/scm_mark0/g
s/\bmark_locations\b/scm_mark_locations/g
s/\bmarkcdr\b/scm_markcdr/g
s/\bmksfpt\b/scm_mksfpt/g
s/\bmkstrport\b/scm_mkstrport/g
s/\bmode_bits\b/scm_mode_bits/g
s/\bmtrigger\b/scm_mtrigger/g
s/\bmust_free\b/scm_must_free/g
s/\bmust_malloc\b/scm_must_malloc/g
s/\bmust_realloc\b/scm_must_realloc/g
s/\bnewptob\b/scm_newptob/g
s/\bnewsmob\b/scm_newsmob/g
s/\bnoop0\b/scm_noop0/g
s/\bnum_finals\b/scm_num_finals/g
s/\bnumptob\b/scm_numptob/g
s/\bnumsmob\b/scm_numsmob/g
s/\bobhash\b/scm_obhash/g
s/\bobunhash\b/scm_obunhash/g
s/\bopen_file\b/scm_open_file/g
s/\boutput_portp\b/scm_output_portp/g
s/\bpipob\b/scm_pipob/g
s/\bprinport\b/scm_prinport/g
s/\bptobs\b/scm_ptobs/g
s/\bs_cells\b/scm_s_cells/g
s/\bs_close_port\b/scm_s_close_port/g
s/\bs_heap\b/scm_s_heap/g
s/\bs_hplims\b/scm_s_hplims/g
s/\bs_nogrow\b/scm_s_nogrow/g
s/\bs_obunhash\b/scm_s_obunhash/g
s/\bsave_regs_gc_mark\b/scm_save_regs_gc_mark/g
s/\bsfflush\b/scm_sfflush/g
s/\bsfwrite\b/scm_sfwrite/g
s/\bsmobs\b/scm_smobs/g
s/\bstack_report\b/scm_stack_report/g
s/\bstrhash\b/scm_strhash/g
s/\bstwrite\b/scm_stwrite/g
s/\bsym2vcell\b/scm_sym2vcell/g
s/\bsymhash_dim\b/scm_symhash_dim/g
s/\bsysintern\b/scm_sysintern/g
s/\bthrowval\b/scm_throwval/g
s/\btmp_errp\b/scm_tmp_errp/g
s/\bcurtime\b/scm_curtime/g
s/\binit_time\b/scm_init_time/g
s/\bmy_base\b/scm_my_base/g
s/\bmy_time\b/scm_my_time/g
s/\btime_in_msec\b/scm_time_in_msec/g
s/\byour_base\b/scm_your_base/g
s/\byour_time\b/scm_your_time/g
s/\barray_equal\b/scm_array_equal/g
s/\binit_unif\b/scm_init_unif/g
s/\bistr2bve\b/scm_istr2bve/g
s/\braprin1\b/scm_raprin1/g
s/\bresizuve\b/scm_resizuve/g
s/\bs_resizuve\b/scm_s_resizuve/g
s/\binit_unix\b/scm_init_unix/g
s/\bl_acct\b/scm_l_acct/g
s/\bl_lstat\b/scm_l_lstat/g
s/\bl_mknod\b/scm_l_mknod/g
s/\bl_nice\b/scm_l_nice/g
s/\bl_readlink\b/scm_l_readlink/g
s/\bl_symlink\b/scm_l_symlink/g
s/\bl_sync\b/scm_l_sync/g
s/\bcell\b/scm_cell/g
s/\bsubr\b/scm_subr/g
s/\biproc\b/scm_iproc/g
s/\bdsubr\b/scm_dsubr/g
s/\bsmobfuns\b/scm_smobfuns/g
s/\bptobfuns\b/scm_ptobfuns/g
s/\barray\b/scm_array/g
s/\barray_dim\b/scm_array_dim/g
s/\bdblproc\b/scm_dblproc/g
s/\bflo\b/scm_flo/g
s/\bdbl\b/scm_dbl/g
s/\bmake_gsubr\b/scm_make_gsubr/g
s/\bgsubr_apply\b/scm_gsubr_apply/g
s/\binit_gsubr\b/scm_init_gsubr/g
s/\binit_ramap\b/\bscm_init_ramap\b/g
s/\bscm_main\b/main/g
s/<scm_string.h>/<string.h>/
s/"scm_/"/g
s/"\*scm_/"\*"/g
s/"\#@scm_/"\#@/g
s/\(".*-\)scm_\(.*"\)/\1\2/g
s/type to scm_apply/type to apply/
s/symbol->scm_/symbol->/g
s/number->scm_/number->/g
s/string->scm_/string->/g
s/vector->scm_/vector->/g
s/list->scm_/list->/g
s/array->scm_/array->/g
s/set-scm_pos/set-pos/g
s/file-scm_pos/file-pos/g
s/-scm_vector/-vector/g
