(in-package guile)
(in-module module)
(define (package-load-module package name hint)
  (display "; (tried to load module ") (write name)
  (display " in ") (write package)
  (display ")\n")
  #f)
(define (package-load-interface package name hint)
  (display "; (tried to load interface ") (write name)
  (display " in ") (write package)
  (display ")\n")
  #f)
