

/* This is the editted-by-hand version of scmflags for guile.
 * The makefile copies this onto scmflags.h.
 */

#define CAUTIOUS
#define IO_EXTENSIONS
#define PROMPT "guile> "
#define FLOATS
#define MEMOIZE_LOCALS
#define GC_FREE_SEGMENTS
#define TICKS
#define MEMOIZE_LOCALS
#define SICP
#define CCLO
#define BIGNUMS
#define ENGNOT
#define ARRAYS

