;;; $Id: defmod.scm,v 1.2 1995/01/06 18:10:33 miles Exp $
;;; ----------------------------------------------------------------
;;; defmod.scm -- Default package/module definition
;;; 21 Dec 1994, Miles Bader <miles@eskimo.com>
;;; ----------------------------------------------------------------
;;;
;;; This file defines the `default' package and module, into which files are
;;; initially loaded (up to the point that they use in-package or in-module).
;;; 

(in-package default)
(use-library guile)

(in-module default)			; in the default package
(use-interface guile)

(in-interface #f)			; no default interface here

(define (%not-in-default-module . noise)
  (error "Module operation cannot be used in the default module"))
(define-macro (export-interfaces-to-library . noise) `(%not-in-default-module))
(define-macro (use-library . noise) `(%not-in-default-module))

;; ----------------------------------------------------------------
