/* classes: h_files */

#ifndef LATTEH
#define LATTEH

/*	Copyright (C) 1995 Free Software Foundation, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */
/*  t. lord	Tue Apr  4 19:59:23 1995	*/



#include "scm.h"

#define LATTE_TYPE_DATA(OBJ)	(CAR (OBJ) - 1)
#define LATTE_DATA(OBJ)		(CDR (OBJ))
#define LATTE_DATA2HANDLE(OBJ)	(*((SCM *)(OBJ) - 1))
#define LATTE_TYPE(OBJ)		LATTE_DATA2HANDLE(LATTE_TYPE_DATA(OBJ))

#define LATTEP(OBJ)		(((CAR (obj) & 3) == 1) \
				 && (CAR (LATTE_TYPE_DATA (obj) == 0)))


#ifdef __STDC__

#else /* STDC */

#endif /* STDC */


#endif  /* LATTEH */
