/* This is the C that was used to generate the
 * the times reported in TIMES.
 */

#include <stdio.h>

main(argc, argv)
     int argc;
     char ** argv;
{
  int count;
  int x;
  int t2;
  int t1;
  int t;

  if (argc != 3)
    exit (1);

  x = atoi (argv[1]) - 2;
  count = atoi (argv[2]);

  while (0 < count--)
    {
      int xx;
      xx = x;
      t2 = t1 = 1;
      while (0 < xx)
	{
	  t = t2 + t1;
	  t2 = t1;
	  t1 = t;
	  --xx;
	}
    }
  printf ("%d\n", t1);
}

