;;; Flag values to pass to Tk_DoOneEvent to disable searches
;;; for some kinds of events:
;;;

(define TK_DONT_WAIT		1)
(define TK_X_EVENTS		2)
(define TK_FILE_EVENTS		4)
(define TK_TIMER_EVENTS		8)
(define TK_IDLE_EVENTS		#x10)
(define TK_ALL_EVENTS		#x1e)

(provide 'Gtk)
