# mkListbox3 w
#
# Create a top-level window containing a listbox with a bunch of well-known
# sayings.  The listbox can be scrolled or scanned in two dimensions.
#
# Arguments:
#    w -	Name to use for new top-level window.
#
# @(#) mkListbox3.tcl 1.2 94/12/23 15:52:08

proc mkListbox3 {{w .l3}} {
    catch {destroy $w}
    toplevel $w
    dpos $w
    wm title $w "Listbox Demonstration (well-known sayings)"
    wm iconname $w "Listbox"
    wm minsize $w 1 1
    label $w.msg -font -Adobe-times-medium-r-normal--*-180-*-*-*-*-*-* \
            -wraplength 3.5i -justify left \
	    -text "The listbox below contains a collection of well-known sayings.  You can scan the list using either of the scrollbars or by dragging in the listbox window with button 2 pressed.  Click the \"OK\" button when you're done."
    frame $w.frame -borderwidth 10
    button $w.ok -text OK -command "destroy $w" -width 8
    pack $w.msg -side top
    pack $w.ok -side bottom
    pack $w.frame -side top -expand yes -fill y

    scrollbar $w.frame.yscroll -command "$w.frame.list yview"
    scrollbar $w.frame.xscroll -orient horizontal \
	    -command "$w.frame.list xview"
    listbox $w.frame.list -width 20 -height 10 -setgrid 1 \
	    -yscroll "$w.frame.yscroll set" -xscroll "$w.frame.xscroll set"
    pack $w.frame.yscroll -side right -fill y
    pack $w.frame.xscroll -side bottom -fill x
    pack $w.frame.list -expand yes -fill y

    $w.frame.list insert 0 "Waste not, want not" "Early to bed and early to rise makes a man healthy, wealthy, and wise" "Ask not what your country can do for you, ask what you can do for your country" "I shall return" "NOT" "A picture is worth a thousand words" "User interfaces are hard to build" "Thou shalt not steal" "A penny for your thoughts" "Fool me once, shame on you;  fool me twice, shame on me" "Every cloud has a silver lining" "Where there's smoke there's fire" "It takes one to know one" "Curiosity killed the cat" "Take this job and shove it" "Up a creek without a paddle" "I'm mad as hell and I'm not going to take it any more" "An apple a day keeps the doctor away" "Don't look a gift horse in the mouth"
}
