# This is the Tcl that was used to generate the
# times reported in TIMES.
#

proc fib_loop {n count}  {
    incr n -2
    set saved_n $n
    set tn1 1
    while {0 < $count} {
	incr count -1
	set tn1 1
	set tn2 1
	set n $saved_n
	while {0 < $n} {
	    set tn [expr $tn1+$tn2]
	    set tn2 $tn1
	    set tn1 $tn
	    incr n -1
	}
     }
     return $tn1
}
