#undef YYDEBUG
#define YYDEBUG 1

#ifdef const
#define DEF_CONST 1
#undef const
#else
#define DEF_CONST 0
#endif

#include "scm.h"
#define YYSTYPE SCM

#if DEF_CONST
#define const
#endif

extern SCM ctax_burst_fn;

#undef YY_INPUT
#define YY_INPUT(buf, result, max_size) \
  result = parse_input (buf, max_size);

extern YYSTYPE parse_number ();
