#ifndef __GNUC__
#define _ili_nop ili_nop
#define _ili_push_0 ili_push_0
#define _ili_iconst_m1 ili_iconst_m1
#define _ili_iconst_0 ili_iconst_0
#define _ili_iconst_1 ili_iconst_1
#define _ili_iconst_2 ili_iconst_2
#define _ili_iconst_3 ili_iconst_3
#define _ili_iconst_4 ili_iconst_4
#define _ili_iconst_5 ili_iconst_5
#define _ili_lconst_0 ili_lconst_0
#define _ili_lconst_1 ili_lconst_1
#define _ili_fconst_0 ili_fconst_0
#define _ili_fconst_1 ili_fconst_1
#define _ili_fconst_2 ili_fconst_2
#define _ili_dconst_0 ili_dconst_0
#define _ili_dconst_1 ili_dconst_1
#define _ili_push_literal ili_push_literal
#define _ili_push_literal2 ili_push_literal2
#define _ili_push_constant ili_push_constant
#define _ili_push_constant2 ili_push_constant2
#define _ili_iload ili_iload
#define _ili_lload ili_lload
#define _ili_iload_0 ili_iload_0
#define _ili_iload_1 ili_iload_1
#define _ili_iload_2 ili_iload_2
#define _ili_iload_3 ili_iload_3
#define _ili_lload_0 ili_lload_0
#define _ili_lload_1 ili_lload_1
#define _ili_lload_2 ili_lload_2
#define _ili_lload_3 ili_lload_3
#define _ili_istore ili_istore
#define _ili_lstore ili_lstore
#define _ili_istore_0 ili_istore_0
#define _ili_istore_1 ili_istore_1
#define _ili_istore_2 ili_istore_2
#define _ili_istore_3 ili_istore_3
#define _ili_lstore_0 ili_lstore_0
#define _ili_lstore_1 ili_lstore_1
#define _ili_lstore_2 ili_lstore_2
#define _ili_lstore_3 ili_lstore_3
#define _ili_iinc ili_iinc
#define _ili_pop ili_pop
#define _ili_pop2 ili_pop2
#define _ili_dup ili_dup
#define _ili_dup2 ili_dup2
#define _ili_dup_x1 ili_dup_x1
#define _ili_dup_x2 ili_dup_x2
#define _ili_dup2_x1 ili_dup2_x1
#define _ili_dup2_x2 ili_dup2_x2
#define _ili_swap ili_swap
#define _ili_lshr ili_lshr
#define _ili_lshl ili_lshl
#define _ili_lushr ili_lushr
#define _ili_land ili_land
#define _ili_lxor ili_lxor
#define _ili_lor ili_lor
#define _ili_ishr ili_ishr
#define _ili_ishl ili_ishl
#define _ili_iushr ili_iushr
#define _ili_iand ili_iand
#define _ili_ixor ili_ixor
#define _ili_ior ili_ior
#define _ili_ldiv ili_ldiv
#define _ili_lmul ili_lmul
#define _ili_ladd ili_ladd
#define _ili_lsub ili_lsub
#define _ili_lmod ili_lmod
#define _ili_idiv ili_idiv
#define _ili_imul ili_imul
#define _ili_iadd ili_iadd
#define _ili_isub ili_isub
#define _ili_imod ili_imod
#define _ili_fdiv ili_fdiv
#define _ili_fmul ili_fmul
#define _ili_fadd ili_fadd
#define _ili_fsub ili_fsub
#define _ili_fmod ili_fmod
#define _ili_ddiv ili_ddiv
#define _ili_dmul ili_dmul
#define _ili_dadd ili_dadd
#define _ili_dsub ili_dsub
#define _ili_dmod ili_dmod
#define _ili_ineg ili_ineg
#define _ili_lneg ili_lneg
#define _ili_dneg ili_dneg
#define _ili_fneg ili_fneg
#define _ili_i2f ili_i2f
#define _ili_i2l ili_i2l
#define _ili_i2d ili_i2d
#define _ili_f2i ili_f2i
#define _ili_f2l ili_f2l
#define _ili_f2d ili_f2d
#define _ili_l2i ili_l2i
#define _ili_l2f ili_l2f
#define _ili_l2d ili_l2d
#define _ili_d2i ili_d2i
#define _ili_d2f ili_d2f
#define _ili_d2l ili_d2l
#define _ili_int2byte ili_int2byte
#define _ili_int2char ili_int2char
#define _ili_int2short ili_int2short
#define _ili_goto ili_goto
#define _ili_jsr ili_jsr
#define _ili_ret ili_ret
#define _ili_tableswitch ili_tableswitch
#define _ili_lookupswitch ili_lookupswitch
#define _ili_icmp ili_icmp
#define _ili_lcmp ili_lcmp
#define _ili_fcmpl ili_fcmpl
#define _ili_dcmpl ili_dcmpl
#define _ili_fcmpg ili_fcmpg
#define _ili_dcmpg ili_dcmpg
#define _ili_ifeq ili_ifeq
#define _ili_ifne ili_ifne
#define _ili_iflt ili_iflt
#define _ili_ifgt ili_ifgt
#define _ili_ifge ili_ifge
#define _ili_ifle ili_ifle
#define _ili_if_icmpeq ili_if_icmpeq
#define _ili_if_icmpne ili_if_icmpne
#define _ili_if_icmplt ili_if_icmplt
#define _ili_if_icmpgt ili_if_icmpgt
#define _ili_if_icmple ili_if_icmple
#define _ili_if_icmpge ili_if_icmpge
#define _ili_newfromname ili_newfromname
#define _ili_new ili_new
#define _ili_ds_new ili_ds_new
#define _ili_getfield ili_getfield
#define _ili_ds_getfield ili_ds_getfield
#define _ili_ds_lgetfield ili_ds_lgetfield
#define _ili_putfield ili_putfield
#define _ili_ds_putfield ili_ds_putfield
#define _ili_ds_lputfield ili_ds_lputfield
#define _ili_getstatic ili_getstatic
#define _ili_ds_getstatic ili_ds_getstatic
#define _ili_ds_lgetstatic ili_ds_lgetstatic
#define _ili_putstatic ili_putstatic
#define _ili_ds_putstatic ili_ds_putstatic
#define _ili_ds_lputstatic ili_ds_lputstatic
#define _ili_newarray ili_newarray
#define _ili_anewarray ili_anewarray
#define _ili_arraylength ili_arraylength
#define _ili_multianewarray ili_multianewarray
#define _ili_iaload ili_iaload
#define _ili_laload ili_laload
#define _ili_faload ili_faload
#define _ili_daload ili_daload
#define _ili_aaload ili_aaload
#define _ili_baload ili_baload
#define _ili_caload ili_caload
#define _ili_saload ili_saload
#define _ili_iastore ili_iastore
#define _ili_lastore ili_lastore
#define _ili_fastore ili_fastore
#define _ili_dastore ili_dastore
#define _ili_aastore ili_aastore
#define _ili_bastore ili_bastore
#define _ili_castore ili_castore
#define _ili_sastore ili_sastore
#define _ili_invokevirtual ili_invokevirtual
#define _ili_ds_invokevirtual ili_ds_invokevirtual
#define _ili_invokenonvirtual ili_invokenonvirtual
#define _ili_invokestatic ili_invokestatic
#define _ili_ds_invoke_known ili_ds_invoke_known
#define _ili_invokeinterface ili_invokeinterface
#define _ili_ds_invokeinterface ili_ds_invokeinterface
#define _ili_vreturn ili_vreturn
#define _ili_ireturn ili_ireturn
#define _ili_lreturn ili_lreturn
#define _ili_instanceof ili_instanceof
#define _ili_ds_instanceof ili_ds_instanceof
#define _ili_checkcast ili_checkcast
#define _ili_ds_checkcast ili_ds_checkcast
#define _ili_athrow ili_athrow
#define _ili_monitorenter ili_monitorenter
#define _ili_monitorexit ili_monitorexit
#define _ili_verifystack ili_verifystack
#define _ili_breakpoint ili_breakpoint
#define _ili_ibitnot ili_ibitnot
#define _ili_icnot ili_icnot
#define _ili_icand ili_icand
#define _ili_icor ili_icor
#define _ili_halt ili_halt
#endif

static void * ili_instruction_mapping[] =
{
  &&_ili_nop,
  &&_ili_push_0,
  &&_ili_iconst_m1,
  &&_ili_iconst_0,
  &&_ili_iconst_1,
  &&_ili_iconst_2,
  &&_ili_iconst_3,
  &&_ili_iconst_4,
  &&_ili_iconst_5,
  &&_ili_lconst_0,
  &&_ili_lconst_1,
  &&_ili_fconst_0,
  &&_ili_fconst_1,
  &&_ili_fconst_2,
  &&_ili_dconst_0,
  &&_ili_dconst_1,
  &&_ili_push_literal,
  &&_ili_push_literal2,
  &&_ili_push_constant,
  &&_ili_push_constant2,
  &&_ili_iload,
  &&_ili_lload,
  &&_ili_iload_0,
  &&_ili_iload_1,
  &&_ili_iload_2,
  &&_ili_iload_3,
  &&_ili_lload_0,
  &&_ili_lload_1,
  &&_ili_lload_2,
  &&_ili_lload_3,
  &&_ili_istore,
  &&_ili_lstore,
  &&_ili_istore_0,
  &&_ili_istore_1,
  &&_ili_istore_2,
  &&_ili_istore_3,
  &&_ili_lstore_0,
  &&_ili_lstore_1,
  &&_ili_lstore_2,
  &&_ili_lstore_3,
  &&_ili_iinc,
  &&_ili_pop,
  &&_ili_pop2,
  &&_ili_dup,
  &&_ili_dup2,
  &&_ili_dup_x1,
  &&_ili_dup_x2,
  &&_ili_dup2_x1,
  &&_ili_dup2_x2,
  &&_ili_swap,
  &&_ili_lshr,
  &&_ili_lshl,
  &&_ili_lushr,
  &&_ili_land,
  &&_ili_lxor,
  &&_ili_lor,
  &&_ili_ishr,
  &&_ili_ishl,
  &&_ili_iushr,
  &&_ili_iand,
  &&_ili_ixor,
  &&_ili_ior,
  &&_ili_ldiv,
  &&_ili_lmul,
  &&_ili_ladd,
  &&_ili_lsub,
  &&_ili_lmod,
  &&_ili_idiv,
  &&_ili_imul,
  &&_ili_iadd,
  &&_ili_isub,
  &&_ili_imod,
  &&_ili_fdiv,
  &&_ili_fmul,
  &&_ili_fadd,
  &&_ili_fsub,
  &&_ili_fmod,
  &&_ili_ddiv,
  &&_ili_dmul,
  &&_ili_dadd,
  &&_ili_dsub,
  &&_ili_dmod,
  &&_ili_ineg,
  &&_ili_lneg,
  &&_ili_dneg,
  &&_ili_fneg,
  &&_ili_i2f,
  &&_ili_i2l,
  &&_ili_i2d,
  &&_ili_f2i,
  &&_ili_f2l,
  &&_ili_f2d,
  &&_ili_l2i,
  &&_ili_l2f,
  &&_ili_l2d,
  &&_ili_d2i,
  &&_ili_d2f,
  &&_ili_d2l,
  &&_ili_int2byte,
  &&_ili_int2char,
  &&_ili_int2short,
  &&_ili_goto,
  &&_ili_jsr,
  &&_ili_ret,
  &&_ili_tableswitch,
  &&_ili_lookupswitch,
  &&_ili_icmp,
  &&_ili_lcmp,
  &&_ili_fcmpl,
  &&_ili_dcmpl,
  &&_ili_fcmpg,
  &&_ili_dcmpg,
  &&_ili_ifeq,
  &&_ili_ifne,
  &&_ili_iflt,
  &&_ili_ifgt,
  &&_ili_ifge,
  &&_ili_ifle,
  &&_ili_if_icmpeq,
  &&_ili_if_icmpne,
  &&_ili_if_icmplt,
  &&_ili_if_icmpgt,
  &&_ili_if_icmple,
  &&_ili_if_icmpge,
  &&_ili_newfromname,
  &&_ili_new,
  &&_ili_ds_new,
  &&_ili_getfield,
  &&_ili_ds_getfield,
  &&_ili_ds_lgetfield,
  &&_ili_putfield,
  &&_ili_ds_putfield,
  &&_ili_ds_lputfield,
  &&_ili_getstatic,
  &&_ili_ds_getstatic,
  &&_ili_ds_lgetstatic,
  &&_ili_putstatic,
  &&_ili_ds_putstatic,
  &&_ili_ds_lputstatic,
  &&_ili_newarray,
  &&_ili_anewarray,
  &&_ili_arraylength,
  &&_ili_multianewarray,
  &&_ili_iaload,
  &&_ili_laload,
  &&_ili_faload,
  &&_ili_daload,
  &&_ili_aaload,
  &&_ili_baload,
  &&_ili_caload,
  &&_ili_saload,
  &&_ili_iastore,
  &&_ili_lastore,
  &&_ili_fastore,
  &&_ili_dastore,
  &&_ili_aastore,
  &&_ili_bastore,
  &&_ili_castore,
  &&_ili_sastore,
  &&_ili_invokevirtual,
  &&_ili_ds_invokevirtual,
  &&_ili_invokenonvirtual,
  &&_ili_invokestatic,
  &&_ili_ds_invoke_known,
  &&_ili_invokeinterface,
  &&_ili_ds_invokeinterface,
  &&_ili_vreturn,
  &&_ili_ireturn,
  &&_ili_lreturn,
  &&_ili_instanceof,
  &&_ili_ds_instanceof,
  &&_ili_checkcast,
  &&_ili_ds_checkcast,
  &&_ili_athrow,
  &&_ili_monitorenter,
  &&_ili_monitorexit,
  &&_ili_verifystack,
  &&_ili_breakpoint,
  &&_ili_ibitnot,
  &&_ili_icnot,
  &&_ili_icand,
  &&_ili_icor,
  &&_ili_halt,
};
